# More Information: https://cloud.google.com/cloud-build/docs/api
from collections.abc import Generator
from dataclasses import dataclass, field
from typing import Any
from urllib.parse import urlparse

from google.api_core.exceptions import AlreadyExists
from google.cloud.devtools import cloudbuild_v1

from gcp_pilot import exceptions
from gcp_pilot.base import GoogleCloudPilotAPI

TriggerType = cloudbuild_v1.BuildTrigger
AnyEventType = cloudbuild_v1.GitHubEventsConfig | cloudbuild_v1.RepoSource


@dataclass
class _SubstitutionVariable:
    key: str
    value: Any
    escape_delimiter: str | None = None

    def __post_init__(self):
        self.escape_delimiter = "|" if "," in str(self.value) else None

    @property
    def full_key(self) -> str:
        # Custom substitution variables must be prefixed with an underscore
        return f"_{self.key}"

    def __str__(self) -> str:
        # When used in a template, $_NAME should work, but it requires to be isolated by blank spaces
        # Thus, we use ${_NAME} by default, because it allows merging with other text.
        return "${%s}" % self.full_key

    @property
    def _escape(self):
        # <https://cloud.google.com/run/docs/configuring/environment-variables#setting>
        return f"^{self.escape_delimiter}^" if self.escape_delimiter else ""

    def as_env_var(self, key: str | None = None):
        return f"{self._escape}{key or self.key}={self!s}"


@dataclass
class Substitutions:
    _variables: dict[str, _SubstitutionVariable] = field(default_factory=dict)

    def add(self, **kwargs):
        for key, val in kwargs.items():
            variable = _SubstitutionVariable(key=key.upper(), value=val)
            self._variables[variable.key] = variable

    @property
    def as_dict(self) -> dict[str, str]:
        return {
            variable.full_key: str(variable.value)  # all values must be string or bytes
            for variable in self._variables.values()
        }

    def __getattr__(self, item: str):
        # Useful tool to ease the variable access (eg. substitution.MY_VAR_NAME)
        return self._variables[item.upper()]


class CloudBuild(GoogleCloudPilotAPI):
    _client_class = cloudbuild_v1.CloudBuildClient

    def make_build_step(
        self,
        name: str,
        identifier: str | None = None,
        args: list | None = None,
        env=None,
        entrypoint: str | None = None,
        timeout: int | None = None,
    ) -> cloudbuild_v1.BuildStep:
        return cloudbuild_v1.BuildStep(
            id=identifier,
            name=name,
            args=args,
            env=env,
            entrypoint=entrypoint,
            timeout=self._as_duration(timeout),
        )

    def make_source_repo_event(
        self,
        repo_name: str,
        branch_name: str | None = None,
        tag_name: str | None = None,
        project_id: str | None = None,
    ) -> cloudbuild_v1.RepoSource:
        if not branch_name and not tag_name:
            branch_name = "master"

        params = {}
        if branch_name:
            params["branch_name"] = branch_name
        if tag_name:
            params["tag_name"] = tag_name
        return cloudbuild_v1.RepoSource(
            project_id=project_id or self.project_id,
            repo_name=repo_name,
            **params,
        )

    def make_github_event(
        self,
        url,
        branch_name: str | None = None,
        tag_name: str | None = None,
    ) -> cloudbuild_v1.GitHubEventsConfig:
        if not branch_name and not tag_name:
            branch_name = "master"

        params = {}
        if branch_name:
            params["branch"] = branch_name
        if tag_name:
            params["tag"] = tag_name

        path = urlparse(url).path
        owner, name = path.split("/")[1:]
        return cloudbuild_v1.GitHubEventsConfig(
            owner=owner,
            name=name,
            push=cloudbuild_v1.PushFilter(
                **params,
            ),
        )

    def _make_trigger(
        self,
        name: str,
        description: str,
        steps: list[cloudbuild_v1.BuildStep],
        event: AnyEventType,
        tags: list[str],
        images: list[str] | None = None,
        substitutions: Substitutions | None = None,
        timeout: int | None = None,
        machine_type: str = cloudbuild_v1.BuildOptions.MachineType.UNSPECIFIED,
    ) -> cloudbuild_v1.BuildTrigger:
        def _get_event_param():
            valid_events = {
                "trigger_template": cloudbuild_v1.RepoSource,
                "github": cloudbuild_v1.GitHubEventsConfig,
            }
            for key, klass in valid_events.items():
                if isinstance(event, klass):
                    return key
            raise exceptions.ValidationError(f"Unsupported event type {event.__class__.__name__,}")

        params = {_get_event_param(): event}

        return cloudbuild_v1.BuildTrigger(
            name=name,
            description=description,
            tags=tags,
            build=cloudbuild_v1.Build(
                steps=steps,
                images=images or [],
                timeout=self._as_duration(timeout),
                options=cloudbuild_v1.BuildOptions(
                    machine_type=machine_type,
                ),
            ),
            substitutions=substitutions.as_dict,
            **params,
        )

    def get_trigger(self, trigger_id: str, project_id: str | None = None) -> TriggerType:
        return self.client.get_build_trigger(
            trigger_id=trigger_id,
            project_id=project_id or self.project_id,
        )

    def delete_trigger(self, trigger_id: str, project_id: str | None = None):
        return self.client.delete_build_trigger(
            trigger_id=trigger_id,
            project_id=project_id or self.project_id,
        )

    def create_trigger(
        self,
        name: str,
        description: str,
        event: AnyEventType,
        steps: list[cloudbuild_v1.BuildStep],
        tags: list[str] | None = None,
        project_id: str | None = None,
        images: list[str] | None = None,
        substitutions: Substitutions | None = None,
        timeout: int | None = None,
        machine_type: str | None = None,
    ) -> TriggerType:
        trigger = self._make_trigger(
            name=name,
            description=description,
            event=event,
            steps=steps,
            tags=tags,
            images=images,
            substitutions=substitutions,
            timeout=timeout,
            machine_type=machine_type,
        )

        response = self.client.create_build_trigger(
            trigger=trigger,
            project_id=project_id or self.project_id,
        )
        return response

    def update_trigger(
        self,
        name: str,
        description: str,
        event: AnyEventType,
        steps: list[cloudbuild_v1.BuildStep],
        tags: list[str] | None = None,
        images: list[str] | None = None,
        substitutions: Substitutions | None = None,
        project_id: str | None = None,
        timeout: int | None = None,
        machine_type: str | None = None,
    ) -> TriggerType:
        trigger = self._make_trigger(
            name=name,
            description=description,
            event=event,
            steps=steps,
            tags=tags,
            images=images,
            substitutions=substitutions,
            timeout=timeout,
            machine_type=machine_type,
        )

        response = self.client.update_build_trigger(
            trigger_id=name,
            trigger=trigger,
            project_id=project_id or self.project_id,
        )
        return response

    def create_or_update_trigger(
        self,
        name: str,
        description: str,
        event: AnyEventType,
        steps: list[cloudbuild_v1.BuildStep],
        tags: list[str] | None = None,
        project_id: str | None = None,
        images: list[str] | None = None,
        substitutions: Substitutions | None = None,
        timeout: int | None = None,
        machine_type: str | None = None,
    ) -> TriggerType:
        create_args = dict(
            name=name,
            description=description,
            event=event,
            steps=steps,
            tags=tags,
            project_id=project_id,
            images=images,
            substitutions=substitutions,
            timeout=timeout,
            machine_type=machine_type,
        )

        try:
            return self.create_trigger(**create_args)
        except AlreadyExists:
            return self.update_trigger(**create_args)

    def run_trigger(
        self,
        name: str,
        tag_name: str | None = None,
        branch_name: str | None = None,
        commit_sha: str | None = None,
        project_id: str | None = None,
    ) -> TriggerType:
        params = [tag_name, branch_name, commit_sha]
        if len([x for x in params if x]) != 1:
            raise exceptions.ValidationError("Only one of {tag_name, branch_name, commit_sha} must be provided")

        event = cloudbuild_v1.RepoSource(
            project_id=project_id or self.project_id,
            tag_name=tag_name,
            branch_name=branch_name,
            commit_sha=commit_sha,
        )
        response = self.client.run_build_trigger(
            trigger_id=name,
            source=event,
            project_id=project_id or self.project_id,
        )
        return response

    def get_builds(
        self,
        trigger_id: str | None = None,
        project_id: str | None = None,
        status: str | None = None,
    ) -> Generator[cloudbuild_v1.Build, None, None]:
        # https://cloud.google.com/cloud-build/docs/view-build-results#filtering_build_results_using_queries
        filters = []
        if trigger_id:
            filters.append(f'trigger_id="{trigger_id}"')

        if status:
            filters.append(f'status="{status}"')

        all_builds = self.client.list_builds(
            filter=" AND ".join(filters),
            project_id=project_id or self.project_id,
        )
        yield from all_builds

    def subscribe(
        self,
        subscription_id: str,
        project_id: str | None = None,
        push_to_url: str | None = None,
        use_oidc_auth: bool | None = False,
    ):
        # https://cloud.google.com/cloud-build/docs/subscribe-build-notifications
        try:
            # try to import here to avoid making pubsub a mandatory dependency of CloudBuild
            from gcp_pilot.pubsub import CloudSubscriber
        except ImportError as exc:
            raise ImportError("Add `pubsub` extras dependency in order to use CloudBuild notifications") from exc
        subscriber = CloudSubscriber()
        subscriber.create_subscription(
            topic_id="cloud-builds",  # pre-defined by GCP
            subscription_id=subscription_id,
            project_id=project_id,
            exists_ok=True,
            auto_create_topic=True,
            push_to_url=push_to_url,
            use_oidc_auth=use_oidc_auth,
        )


__all__ = (
    "CloudBuild",
    "Substitutions",
)



import abc
import json
import logging
import os
from collections.abc import Callable, Generator
from functools import cached_property
from typing import Any

import google.auth.transport._http_client
from google import auth
from google.auth import iam
from google.auth.credentials import Credentials
from google.auth.impersonated_credentials import Credentials as ImpersonatedCredentials
from google.auth.transport import requests
from google.oauth2.service_account import Credentials as ServiceAccountCredentials
from google.protobuf.duration_pb2 import Duration
from googleapiclient.discovery import Resource, build
from googleapiclient.errors import HttpError
from requests import HTTPError, Response

from gcp_pilot import exceptions

DEFAULT_PROJECT = os.environ.get("GCP_PROJECT", None)
DEFAULT_LOCATION = os.environ.get("GCP_LOCATION", None)
DEFAULT_SERVICE_ACCOUNT = os.environ.get("GCP_SERVICE_ACCOUNT", None)

TOKEN_URI = "https://accounts.google.com/o/oauth2/token"

PolicyType = dict[str, Any]
AuthType = tuple[Credentials, str, str]
ImpersonatedAuthType = tuple[ImpersonatedCredentials, str]
ResourceType = dict[str, Any]

logger = logging.getLogger()

_CACHED_LOCATIONS = {}  # TODO: Implement a smarter solution for caching project's location


MINIMAL_SCOPES = [
    "https://www.googleapis.com/auth/cloud-platform",
]


class GoogleCloudPilotAPI(abc.ABC):
    _client_class = None
    _scopes: list[str] = []
    _iam_roles: list[str] = []
    _cached_credentials: AuthType | None = None
    _service_name = None
    _google_managed_service = False  # Service agent requires impersonation

    def __init__(
        self,
        subject: str | None = None,
        location: str | None = None,
        project_id: str | None = None,
        impersonate_account: str | None = None,
        credentials: Credentials | None = None,
        **kwargs,
    ):
        if credentials:
            self.credentials = credentials
            credential_project_id = getattr(credentials, "project_id", None)
            self._service_account_email = getattr(credentials, "service_account_email", DEFAULT_SERVICE_ACCOUNT)
        else:
            self.credentials, credential_project_id, self._service_account_email = self._set_credentials(
                subject=subject,
                impersonate_account=impersonate_account,
            )
        self.project_id = self._set_project_id(project_id=project_id, credential_project_id=credential_project_id)

        self.client = self._build_client(**kwargs)

        self._location = location or DEFAULT_LOCATION

    @cached_property
    def service_account_email(self) -> str:
        if self._service_account_email == "default":
            self.credentials.refresh(request=google.auth.transport._http_client.Request())
            self._service_account_email = getattr(
                self.credentials,
                "service_account_email",
                self._service_account_email,
            )
        elif self._service_account_email is None:  # SDK local authentication
            self._service_account_email = DEFAULT_SERVICE_ACCOUNT
            if not self._service_account_email:
                logger.warning(
                    "Using local SDK authentication. Set GCP_SERVICE_ACCOUNT or some features might not work.",
                )
        return self._service_account_email

    def _get_client_extra_kwargs(self):
        return {}

    def _build_client(self, **kwargs) -> Resource | _client_class:
        kwargs.update(self._get_client_extra_kwargs())

        return (self._client_class or build)(credentials=self.credentials, **kwargs)

    def _set_project_id(self, project_id: str, credential_project_id: str) -> str:
        return project_id or DEFAULT_PROJECT or credential_project_id

    def _set_location(self, location: str | None = None) -> str:
        return location or DEFAULT_LOCATION or self._get_project_default_location()

    @property
    def location(self):
        if not self._location:
            self._location = self._get_project_default_location()
        return self._location

    @classmethod
    def _impersonate_account(
        cls,
        credentials: Credentials,
        service_account: str,
        scopes: list[str],
    ) -> ImpersonatedAuthType:
        credentials = ImpersonatedCredentials(
            source_credentials=credentials,
            target_principal=service_account,
            target_scopes=scopes,
            delegates=[],
        )

        # Fetch project from service account
        # Since we are impersonating this service account, use it's own project
        # TODO: regex this
        project_id = credentials.service_account_email.split("@")[-1].replace(".iam.gserviceaccount.com", "")

        return credentials, project_id

    @classmethod
    def _delegated_credential(
        cls,
        credentials: Credentials,
        subject: str,
        scopes: list[str],
    ) -> ServiceAccountCredentials:
        try:
            admin_credentials = credentials.with_subject(subject).with_scopes(scopes)
        except AttributeError:
            # When inside GCP, the credentials provided by the metadata service are immutable.
            # https://github.com/GoogleCloudPlatform/professional-services/tree/master/examples/gce-to-adminsdk

            request = requests.Request()
            credentials.refresh(request)

            # Create an IAM signer using the bootstrap credentials.
            signer = iam.Signer(
                request,
                credentials,
                credentials.service_account_email,
            )
            # Create OAuth 2.0 Service Account credentials using the IAM-based
            # signer and the bootstrap_credential's service account email.
            admin_credentials = ServiceAccountCredentials(
                signer,
                credentials.service_account_email,
                TOKEN_URI,
                scopes=scopes,
                subject=subject,
            )

        return admin_credentials

    @classmethod
    def _set_credentials(cls, subject: str | None = None, impersonate_account: str | None = None) -> AuthType:
        # Speed up consecutive authentications
        # TODO: check if this does not break multiple client usage
        all_scopes = MINIMAL_SCOPES + cls._scopes
        if not cls._cached_credentials:
            credentials, project_id = auth.default(scopes=all_scopes)
            current_account = getattr(credentials, "service_account_email", None)
            cls._cached_credentials = credentials, project_id, current_account
        else:
            credentials, project_id, current_account = cls._cached_credentials

        if current_account == "default":  # common when inside GCP
            current_account = DEFAULT_SERVICE_ACCOUNT
        elif current_account is None:  # common when local SDK authentication
            impersonate_account = DEFAULT_SERVICE_ACCOUNT

        if impersonate_account and impersonate_account != current_account:
            credentials, impersonated_project_id = cls._impersonate_account(
                credentials=credentials,
                service_account=impersonate_account,
                scopes=all_scopes,
            )
            project_id = impersonated_project_id or project_id
            service_account = impersonate_account
        else:
            service_account = current_account

        if subject:
            credentials = cls._delegated_credential(credentials=credentials, subject=subject, scopes=all_scopes)

        return credentials, (project_id or getattr(credentials, "project_id", None)), service_account

    def get_oidc_token(self, audience: str | None = None) -> dict[str, dict[str, str]]:
        oidc_token = {"service_account_email": self.service_account_email}
        if audience:
            # TODO: make sure that, if URL, the query params are removed
            oidc_token["audience"] = audience
        return {"oidc_token": oidc_token}

    def set_up_permissions(self, email: str, project_id: str | None = None) -> None:
        from gcp_pilot.resource import ResourceManager, ServiceAgent

        rm = ResourceManager()
        for role in self._iam_roles:
            rm.add_member(
                email=email,
                role=role,
                project_id=project_id or self.project_id,
            )

        if self._google_managed_service:
            email = ServiceAgent.get_email(
                service_name=f"{self._service_name} Service Account",
                project_id=self.project_id,
            )

            ResourceManager().allow_impersonation(
                email=email,
                project_id=project_id,
            )

    def _get_project_number(self, project_id: str) -> int:
        from gcp_pilot.resource import ResourceManager

        project = ResourceManager().get_project(project_id=project_id)
        return project["projectNumber"]

    def _as_duration(self, seconds) -> Duration:
        return Duration(seconds=seconds) if seconds else None

    @classmethod
    def build_from(cls, client: "GoogleCloudPilotAPI", project_id: str | None = None):
        return cls(
            credentials=client.credentials,
            project_id=project_id or client.project_id,
        )

    def _get_project_default_location(self, project_id: str | None = None) -> str | None:
        location = _CACHED_LOCATIONS.get(project_id or self.project_id, None)
        if location:
            return location

        from gcp_pilot.app_engine import AppEngine

        try:
            app_engine = AppEngine.build_from(client=self, project_id=project_id)
        except exceptions.NotFound:
            return None
        return app_engine.location

    def _project_path(self, project_id: str | None = None) -> str:
        return f"projects/{project_id or self.project_id}"

    def _location_path(self, project_id: str | None = None, location: str | None = None) -> str:
        project_path = self._project_path(project_id=project_id)
        return f"{project_path}/locations/{location or self.location}"

    @property
    def _session(self) -> requests.AuthorizedSession:
        return requests.AuthorizedSession(credentials=self.credentials)

    @property
    def _base_url(self) -> str:
        metadata = self.client._rootDesc
        return f"{metadata['baseUrl']}{metadata['version']}"


class AccountManagerMixin:
    def _as_member(self, email: str) -> str:
        if email == "allUsers":
            return email
        is_service_account = email.endswith(".gserviceaccount.com")
        prefix = "serviceAccount" if is_service_account else "member"
        return f"{prefix}:{email}"

    def _make_public(self, role: str, policy: dict) -> dict:
        return self._bind_email_to_policy(email="allUsers", role=role, policy=policy)

    def _make_private(self, role: str, policy: dict) -> dict:
        return self._unbind_email_from_policy(email="allUsers", role=role, policy=policy)

    def _bind_email_to_policy(self, email: str, role: str, policy: dict) -> dict:
        new_policy = policy.copy()

        role_id = role if (role.startswith("organizations/") or role.startswith("roles/")) else f"roles/{role}"
        member = self._as_member(email=email)

        try:
            binding = next(b for b in new_policy["bindings"] if b["role"] == role_id)
            if member in binding["members"]:
                return new_policy
            binding["members"].append(member)
        except (StopIteration, KeyError):
            binding = {"role": role_id, "members": [member]}

            new_bindings = new_policy.get("bindings", []).copy()
            new_bindings.append(binding)

            new_policy["bindings"] = new_bindings

        if "version" not in new_policy:
            new_policy["version"] = 1  # TODO: handle version 2 and 3 as its conditional roles
        return new_policy

    def _unbind_email_from_policy(self, email: str, role: str, policy: dict):
        role_id = f"roles/{role}"
        member = self._as_member(email=email)

        try:
            binding = next(b for b in policy["bindings"] if b["role"] == role_id)
            if member not in binding["members"]:
                return policy
            binding["members"].remove(member)
        except (StopIteration, KeyError):
            pass
        return policy


class AppEngineBasedService:
    # This mixin must be used for all clients that uses GCP's resources that are based on App Engine
    # such as Cloud Scheduler.
    # They have a peculiar behaviour because the App Engine, when enabled the first time,
    # is created in a region and this region cannot be changed ever.
    # So these clients cannot just choose a desired region to work on, they must use the App Engine's
    # previously chosen region.
    def _set_location(self, location: str | None = None):
        project_location = self._get_project_default_location()

        explicit_location = location
        if explicit_location and explicit_location != project_location:
            logger.warning(
                f"Location {explicit_location} cannot be set in {self.__class__.__name__}. "
                f"It uses App Engine's default location for your project: {project_location}",
            )
        return project_location


def friendly_http_error(func):
    _reasons = {
        "notFound": exceptions.NotFound,
        "deleted": exceptions.AlreadyDeleted,
        "forbidden": exceptions.NotAllowed,
        "duplicate": exceptions.AlreadyExists,
        "push.webhookUrlUnauthorized": exceptions.PushWebhookInvalid,
        "channelIdNotUnique": exceptions.ChannelIdNotUnique,
        "notACalendarUser": exceptions.NotACalendarUser,
        "quotaExceeded": exceptions.QuotaExceeded,
        "invalid": exceptions.ValidationError,
    }
    _statuses = {
        "INVALID_ARGUMENT": exceptions.ValidationError,
        "PERMISSION_DENIED": exceptions.NotAllowed,
        "NOT_FOUND": exceptions.NotFound,
        "ALREADY_EXISTS": exceptions.AlreadyExists,
        "INVALID_PASSWORD": exceptions.InvalidPassword,
        "EMAIL_NOT_FOUND": exceptions.NotFound,
        "MISSING_ID_TOKEN": exceptions.MissingUserIdentification,
        "FAILED_PRECONDITION": exceptions.FailedPrecondition,
    }

    def inner_function(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (HttpError, HTTPError) as exc:
            error_content = json.loads(exc.content) if isinstance(exc, HttpError) else exc.response.json()
            if "issue" in error_content:
                raise exceptions.OperationError(errors=error_content["issue"]) from exc

            errors = error_content["error"]
            exception_klass = None
            details = ""

            if "errors" in errors:
                main_error = errors["errors"][0]["reason"]
                exception_klass = _reasons.get(main_error, None)
                details = errors.get("message", "")

            if not exception_klass and "message" in errors:
                exception_klass = _statuses.get(errors["message"], None)

            if not exception_klass and "status" in errors:
                exception_klass = _statuses.get(errors["status"], None)
                details = f"{errors['code']}: {errors['message']}"

            if exception_klass:
                raise exception_klass(details) from exc
            raise

    return inner_function


class DiscoveryMixin:
    @friendly_http_error
    def _execute(self, method: Callable, method_http_headers=None, **kwargs) -> ResourceType:
        call = method(**kwargs)
        if isinstance(call, Response):
            call.raise_for_status()
            return call.json()
        if method_http_headers:
            call.headers = (call.headers or {}) | method_http_headers
        return call.execute()

    def _list(
        self,
        method: Callable,
        result_key: str = "items",
        params: dict[str, Any] | None = None,
    ) -> Generator[ResourceType, None, None]:
        results = self._execute(
            method=method,
            **params,
        )
        yield from results.get(result_key, [])

    def _paginate(
        self,
        method: Callable,
        result_key: str = "items",
        pagination_key: str = "pageToken",  # Because GCP can be weird sometimes
        next_pagination_key: str = "nextPageToken",
        params: dict[str, Any] | None = None,
        order_by: str | None = None,
        limit: int | None = None,
    ) -> Generator[ResourceType, None, None]:
        page_token = None
        params = params or {}

        if order_by:
            if order_by.startswith("-"):
                params["sortOrder"] = "DESCENDING"
                order_by = order_by[1:]
            params["orderBy"] = order_by

        if limit:
            params["maxResults"] = limit

        while True:
            call_kwargs = params.copy()
            if page_token:
                call_kwargs[pagination_key] = page_token
            results = self._execute(
                method=method,
                **call_kwargs,
            )
            yield from results.get(result_key, [])

            page_token = results.get(next_pagination_key)
            if not page_token:
                break



# More Information: <https://cloud.google.com/iam/docs/reference/rest>
import base64
import json
from collections.abc import Generator
from datetime import UTC, datetime, timedelta
from typing import Any

import requests
from google.auth import jwt
from google.cloud import iam_credentials_v1

from gcp_pilot import exceptions
from gcp_pilot.base import AccountManagerMixin, DiscoveryMixin, GoogleCloudPilotAPI, PolicyType, friendly_http_error

AccountType = dict[str, Any]
KeyType = dict[str, Any]


IDP_JWT_AUDIENCE = "https://identitytoolkit.googleapis.com/google.identity.identitytoolkit.v1.IdentityToolkit"


class IdentityAccessManager(AccountManagerMixin, DiscoveryMixin, GoogleCloudPilotAPI):
    def __init__(self, **kwargs):
        super().__init__(
            serviceName="iam",
            version="v1",
            cache_discovery=False,
            **kwargs,
        )

    def _service_account_path(self, email: str, project_id: str | None = None) -> str:
        parent_path = self._project_path(project_id=project_id)
        return f"{parent_path}/serviceAccounts/{email}"

    def _key_path(self, key_id: str, email: str, project_id: str | None = None) -> str:
        parent_path = self._service_account_path(email=email, project_id=project_id)
        return f"{parent_path}/keys/{key_id}"

    def _build_service_account_email(self, name: str, project_id: str | None = None) -> str:
        return f"{name}@{project_id or self.project_id}.iam.gserviceaccount.com"

    def get_service_account(self, name: str, project_id: str | None = None) -> AccountType:
        account_path = self._service_account_path(
            email=self._build_service_account_email(name=name, project_id=project_id),
            project_id=project_id,
        )

        return self._execute(
            method=self.client.projects().serviceAccounts().get,
            name=account_path,
        )

    def create_service_account(
        self,
        name: str,
        display_name: str,
        project_id: str | None = None,
        exists_ok: bool = True,
    ) -> AccountType:
        try:
            body = {"accountId": name, "serviceAccount": {"displayName": display_name}}
            service_account = self._execute(
                method=self.client.projects().serviceAccounts().create,
                name=self._project_path(project_id=project_id),
                body=body,
            )
        except exceptions.AlreadyExists:
            if not exists_ok:
                raise
            service_account = self.get_service_account(name=name, project_id=project_id)
        return service_account

    def list_service_accounts(self, project_id: str | None = None) -> Generator[AccountType, None, None]:
        params = dict(
            name=self._project_path(project_id=project_id),
        )
        pagination = self._paginate(
            method=self.client.projects().serviceAccounts().list,
            result_key="accounts",
            params=params,
        )
        yield from pagination

    def get_policy(self, email: str, project_id: str | None = None) -> PolicyType:
        resource = self._service_account_path(email=email, project_id=project_id)
        return self._execute(
            method=self.client.projects().serviceAccounts().getIamPolicy,
            resource=resource,
        )

    def _as_member(self, email: str) -> str:
        is_service_account = email.endswith(".gserviceaccount.com")
        prefix = "serviceAccount" if is_service_account else "member"
        return f"{prefix}:{email}"

    def bind_member(self, target_email: str, member_email: str, role: str, project_id=None) -> PolicyType:
        policy = self.get_policy(email=target_email, project_id=project_id)
        changed_policy = self._bind_email_to_policy(email=member_email, role=role, policy=policy)
        return self.set_policy(email=target_email, policy=changed_policy, project_id=project_id)

    def remove_member(
        self,
        target_email: str,
        member_email: str,
        role: str,
        project_id: str | None = None,
    ) -> PolicyType:
        policy = self.get_policy(email=target_email, project_id=project_id)
        changed_policy = self._unbind_email_from_policy(email=member_email, role=role, policy=policy)
        return self.set_policy(email=target_email, policy=changed_policy, project_id=project_id)

    def set_policy(self, email: str, policy: PolicyType, project_id: str | None = None) -> PolicyType:
        resource = self._service_account_path(email=email, project_id=project_id)
        return self._execute(
            method=self.client.projects().serviceAccounts().setIamPolicy,
            resource=resource,
            body={"policy": policy, "updateMask": "bindings"},
        )

    def get_key(self, key_id: str, service_account_name: str, project_id: str | None = None) -> KeyType:
        key_path = self._key_path(
            key_id=key_id,
            email=self._build_service_account_email(name=service_account_name, project_id=project_id),
            project_id=project_id,
        )

        return self._execute(
            method=self.client.projects().serviceAccounts().key().get,
            name=key_path,
        )

    def delete_key(self, key_id: str, service_account_name: str, project_id: str | None = None) -> KeyType:
        key_path = self._key_path(
            key_id=key_id,
            email=self._build_service_account_email(name=service_account_name, project_id=project_id),
            project_id=project_id,
        )

        return self._execute(
            method=self.client.projects().serviceAccounts().keys().delete,
            name=key_path,
        )

    def create_key(
        self,
        service_account_name: str,
        project_id: str | None = None,
    ) -> KeyType:
        body = {}
        parent = self._service_account_path(
            email=self._build_service_account_email(name=service_account_name, project_id=project_id),
            project_id=project_id,
        )
        account_key = self._execute(
            method=self.client.projects().serviceAccounts().keys().create,
            name=parent,
            body=body,
        )
        return self._format_key(data=account_key)

    def list_keys(self, service_account_name: str, project_id: str | None = None) -> Generator[KeyType, None, None]:
        parent = self._service_account_path(
            email=self._build_service_account_email(name=service_account_name, project_id=project_id),
            project_id=project_id,
        )
        params = dict(
            name=parent,
        )
        pagination = self._list(
            method=self.client.projects().serviceAccounts().keys().list,
            result_key="keys",
            params=params,
        )
        for item in pagination:
            yield self._format_key(data=item)

    def _format_key(self, data: dict) -> dict:
        prefix, suffix = data["name"].split("/keys/", 1)
        data["id"] = suffix
        data["service_account_email"] = prefix.rsplit("/", 1)[-1]
        data["service_account_name"] = data["service_account_email"].rsplit("@", 1)[0]

        if "privateKeyData" in data:
            data["json"] = base64.b64decode(data["privateKeyData"]).decode()
        return data


class IAMCredentials(GoogleCloudPilotAPI):
    _client_class = iam_credentials_v1.IAMCredentialsClient

    @friendly_http_error
    def encode_jwt(self, payload: dict, service_account_email: str | None) -> str:
        max_expiration = 12 * 60 * 60
        if "iat" not in payload:
            payload["iat"] = datetime.now(tz=UTC).timestamp()
        if "exp" not in payload:
            payload["exp"] = datetime.now(tz=UTC).timestamp() + max_expiration
        elif payload["exp"] - datetime.now(tz=UTC).timestamp() > max_expiration:
            raise ValueError("JWT tokens cannot be valid for more than 12 hours")

        payload["iam"] = int(payload["iat"])
        payload["exp"] = int(payload["exp"])

        response = self.client.sign_jwt(
            name=self.client.service_account_path(service_account=service_account_email, project="-"),
            payload=json.dumps(payload),
        )
        return response.signed_jwt

    @friendly_http_error
    def generate_id_token(self, audience: str, service_account_email: str | None = None) -> str:
        response = self.client.generate_id_token(
            name=self.client.service_account_path(
                service_account=service_account_email or self.service_account_email,
                project="-",
            ),
            audience=audience,
        )
        return response.token

    @classmethod
    def decode_jwt(cls, token: str, issuer_email: str, audience: str | None, verify: bool = True) -> dict[str, Any]:
        certs = cls._fetch_public_certs(email=issuer_email)
        return dict(
            jwt.decode(
                token=token,
                certs=certs,
                audience=audience,
                verify=verify,
            ),
        )

    @classmethod
    def decode_id_token(cls, token: str, issuer_email: str, verify: bool = True) -> dict[str, Any]:
        return cls.decode_jwt(
            token=token,
            issuer_email=issuer_email,
            audience=IDP_JWT_AUDIENCE,
            verify=verify,
        )

    def generate_custom_token(
        self,
        uid: str,
        expires_in_seconds: int,
        tenant_id: str | None = None,
        auth_email: str | None = None,
        claims: dict | None = None,
    ) -> str:
        authenticator_email = auth_email or self.service_account_email
        payload = {
            "iat": datetime.now(tz=UTC).timestamp(),
            "exp": (datetime.now(tz=UTC) + timedelta(seconds=expires_in_seconds)).timestamp(),
            "aud": IDP_JWT_AUDIENCE,
            "iss": authenticator_email,
            "sub": authenticator_email,
            "email": authenticator_email,
            "uid": uid,
            "tenant_id": tenant_id,
            "claims": claims or {},
        }

        return self.encode_jwt(
            payload=payload,
            service_account_email=authenticator_email,
        )

    @classmethod
    def _fetch_public_certs(cls, email: str) -> dict:
        url = f"https://www.googleapis.com/robot/v1/metadata/x509/{email}"
        response = requests.get(url=url, timeout=5)
        return response.json()


__all__ = (
    "IdentityAccessManager",
    "IAMCredentials",
)



# More Information: https://developers.google.com/calendar/v3/reference
import datetime
from collections.abc import Generator
from dataclasses import dataclass
from enum import Enum
from uuid import uuid4

import pytz

from gcp_pilot.base import DiscoveryMixin, GoogleCloudPilotAPI, ResourceType


class ResponseStatus(Enum):
    # Predefined options to attendee response status
    # Refs: <https://developers.google.com/calendar/api/v3/reference/events/update#attendees.responseStatus>

    needs_action = "needsAction"
    declined = "declined"
    tentative = "tentative"
    accepted = "accepted"


@dataclass
class Attendee:
    email: str
    optional: bool = False
    response_status: ResponseStatus | None = None

    def as_data(self) -> dict:
        return {
            "email": self.email,
            "optional": self.optional,
            "responseStatus": self.response_status,
        }


class Color(Enum):
    # Predefined colors.
    # For more color, use the endpoint: <https://developers.google.com/calendar/v3/reference/colors>

    light_blue = ("1", "#a4bdfc")
    light_green = ("2", "#7ae7bf")
    light_purple = ("3", "#dbadff")
    light_red = ("4", "#ff887c")
    yellow = ("5", "#fbd75b")
    light_orange = ("6", "#ffb878")
    water_green = ("7", "#46d6db")
    gray = ("8", "#e1e1e1")
    blue = ("9", "#5484ed")
    green = ("10", "#51b749")
    red = ("11", "#dc2127")


class SendUpdates(Enum):
    # Predefined options to send updates on events.
    # Refs: <https://developers.google.com/calendar/api/v3/reference/events/update#sendUpdates>

    all = "all"
    external_only = "externalOnly"
    none = "none"


class Calendar(DiscoveryMixin, GoogleCloudPilotAPI):
    _scopes = ["https://www.googleapis.com/auth/calendar"]

    def __init__(self, email: str, timezone: str = "UTC", **kwargs):
        self.email = email
        self.timezone = pytz.timezone(timezone)

        super().__init__(
            serviceName="calendar",
            version="v3",
            cache_discovery=False,
            subject=email,
            **kwargs,
        )

    def _date_to_str(self, dt: datetime.date, fmt="%Y-%m-%dT%H:%M:%SZ"):
        if isinstance(dt, datetime.datetime):
            if not dt.tzinfo:
                dt = self.timezone.localize(dt)
            dt = dt.astimezone(self.timezone)
        elif isinstance(dt, datetime.date):
            dt = datetime.datetime(year=dt.year, month=dt.month, day=dt.day, tzinfo=self.timezone)

        return dt.strftime(fmt)

    def get_calendars(self) -> Generator[ResourceType, None, None]:
        params = {}
        yield from self._paginate(
            method=self.client.calendarList().list,
            result_key="items",
            params=params,
        )

    def get_calendar(self, calendar_id: str = "primary") -> ResourceType:
        return self._execute(
            method=self.client.calendars().get,
            calendarId=calendar_id,
        )

    def clear_calendar(self, calendar_id: str = "primary") -> ResourceType:
        params = {}
        return self._execute(
            method=self.client.calendars().clear,
            calendarId=calendar_id,
            params=params,
        )

    def delete_calendar(self, calendar_id: str = "primary") -> ResourceType:
        params = {}
        return self._execute(
            method=self.client.calendars().delete,
            calendarId=calendar_id,
            params=params,
        )

    def create_or_update_calendar(
        self,
        summary: str,
        description: str = "",
        timezone: str | None = None,
        calendar_id: str | None = None,
    ) -> ResourceType:
        if not calendar_id:
            return self.create_calendar(summary=summary, description=description, timezone=timezone)
        return self.update_calendar(
            calendar_id=calendar_id,
            summary=summary,
            description=description,
            timezone=timezone,
        )

    def create_calendar(self, summary: str, description: str = "", timezone: str | None = None) -> ResourceType:
        data = {
            "summary": summary,
            "description": description,
            "timeZone": timezone or self.timezone,
        }
        return self._execute(
            method=self.client.calendars().insert,
            body=data,
        )

    def update_calendar(
        self,
        calendar_id: str,
        summary: str | None = None,
        description: str | None = None,
        timezone: str | None = None,
    ) -> ResourceType:
        data = {}
        if summary:
            data["summary"] = summary
        if description:
            data["description"] = description
        if timezone:
            data["timeZone"] = timezone or self.timezone

        return self._execute(
            method=self.client.calendars().update,
            calendarId=calendar_id,
            body=data,
        )

    def watch_calendars(self, hook_url: str, hook_token: str | None = None, uuid: str | None = None) -> ResourceType:
        data = {
            "id": uuid or uuid4().hex,
            "type": "webhook",
            "address": hook_url,
        }
        if hook_token:
            data["token"] = hook_token

        return self._execute(
            method=self.client.calendarList().watch,
            body=data,
        )

    def unwatch(self, uuid: str, resource_id: str) -> ResourceType:
        data = {
            "id": uuid,
            "resource_id": resource_id,
        }
        return self._execute(
            method=self.client.channels().stop,
            body=data,
        )

    def create_or_update_event(
        self,
        summary: str,
        start_at: datetime.date,
        end_at: datetime.date,
        location: str | None = None,
        event_id: str | None = None,
        description: str | None = None,
        attendees: list[Attendee] | None = None,
        recurrence_type: str | None = None,
        recurrence_amount: str | None = None,
        calendar_id: str | None = "primary",
        color: Color | None = None,
        send_updates: SendUpdates | None = None,
    ) -> ResourceType:
        def _build_time_field(dt):
            if isinstance(start_at, datetime.datetime):
                return {
                    "dateTime": self._date_to_str(dt),
                }

            return {
                "date": self._date_to_str(dt, fmt="%Y-%m-%d"),
            }

        data = {
            "summary": summary,
            "location": location,
            "description": description,
            "start": _build_time_field(dt=start_at),
            "end": _build_time_field(dt=end_at),
            "sendUpdates": send_updates,
        }

        if color:
            data["colorId"] = color[0]

        if recurrence_type and recurrence_amount:
            data["recurrence"] = [f"RRULE:FREQ={recurrence_type};COUNT={recurrence_amount}"]

        if attendees:
            data["attendees"] = [attendee.as_data() for attendee in attendees]

        if event_id:
            return self._execute(
                method=self.client.events().update,
                calendarId=calendar_id,
                eventId=event_id,
                body=data,
            )

        return self._execute(
            method=self.client.events().insert,
            calendarId=calendar_id,
            body=data,
        )

    def get_events(
        self,
        calendar_id: str = "primary",
        starts_at: datetime.date | None = None,
        ends_at: datetime.date | None = None,
    ) -> Generator[ResourceType, None, None]:
        min_date = self._date_to_str(starts_at) if starts_at else None
        max_date = self._date_to_str(ends_at) if ends_at else None

        page_size = 100
        params = dict(
            calendarId=calendar_id,
            timeMin=min_date,
            timeMax=max_date,
            singleEvents=True,
        )

        yield from self._paginate(
            method=self.client.events().list,
            result_key="items",
            params=params,
            order_by="startTime",
            limit=page_size,
        )

    def get_event(self, event_id: str, calendar_id: str = "primary") -> ResourceType:
        return self._execute(
            method=self.client.events().get,
            calendarId=calendar_id,
            eventId=event_id,
        )

    def delete_event(self, event_id: str, calendar_id: str = "primary") -> ResourceType:
        return self._execute(
            method=self.client.events().delete,
            calendarId=calendar_id,
            eventId=event_id,
        )

    def get_recurrent_events(self, event_id: str, calendar_id: str = "primary") -> Generator[ResourceType, None, None]:
        page_size = 100
        params = dict(
            calendarId=calendar_id,
            eventId=event_id,
        )

        yield from self._paginate(
            method=self.client.events().instances,
            result_key="items",
            params=params,
            limit=page_size,
        )

    def watch_events(
        self,
        hook_url: str,
        calendar_id: str = "primary",
        hook_token: str | None = None,
        uuid: str | None = None,
    ) -> ResourceType:
        data = {
            "id": uuid or uuid4().hex,
            "type": "webhook",
            "address": hook_url,
        }
        if hook_token:
            data["token"] = hook_token

        return self._execute(
            method=self.client.events().watch,
            calendarId=calendar_id,
            body=data,
        )

    def check_availability(
        self,
        starts_at: datetime,
        ends_at: datetime,
        timezone: str = None,
        calendar_ids: list[str] | None = None,
        calendar_id: str = "primary",
    ) -> ResourceType:
        data = {
            "timeMin": self._date_to_str(starts_at),
            "timeMax": self._date_to_str(ends_at),
            "timeZone": timezone or self.timezone,
            "groupExpansionMax": 100,  # max value
            "calendarExpansionMax": 50,  # max value
            "items": [{"id": calendar_id} for calendar_id in (calendar_ids or ["primary"])],
        }

        return self._execute(
            method=self.client.freeBusy().watch,
            calendarId=calendar_id,
            body=data,
        )["calendars"]


__all__ = (
    "Calendar",
    "Attendee",
    "Color",
    "ResponseStatus",
    "SendUpdates",
)



import json
from collections.abc import Generator
from dataclasses import dataclass, field

import requests

from gcp_pilot import exceptions
from gcp_pilot.base import DiscoveryMixin, GoogleCloudPilotAPI, ResourceType


class Text:
    @classmethod
    def build_mention(cls, member_id: str = "all") -> str:
        return f"<users/{member_id}>"

    @classmethod
    def build_link(cls, url: str, text: str) -> str:
        return f"<{url}|{text}>"

    @classmethod
    def format_color(cls, hex_color: str, text: str) -> str:
        return f'<font color="#{hex_color}">{text}</font>'


class Widget(dict):
    _key: str | None = None

    def as_data(self):
        if self._key:
            return {self._key: dict(self)}

        return dict(self)


class ButtonWidget(Widget):
    def __init__(self, url, text: str | None = None, image_url: str | None = None, icon: str | None = None):
        super().__init__()
        if text:
            self._key = "textButton"
            self["text"] = text
        elif image_url:
            self._key = "imageButton"
            self["iconUrl"] = image_url
        elif icon:
            self._key = "imageButton"
            self["icon"] = icon
        else:
            raise exceptions.UnsupportedFormatException("A button must have a text, image or icon")

        self["onClick"] = {
            "openLink": {
                "url": url,
            },
        }


class ButtonGroupWidget(Widget):
    _key = "buttons"

    def __init__(self, buttons):
        data = dict(
            buttons=buttons,
        )
        super().__init__(data)

    def as_data(self):
        return {self._key: [button.as_data() for button in self["buttons"]]}


class OnClickWidget(Widget):
    _key = "onClick"

    def __init__(self, url):
        data = dict(openLink=dict(url=url))
        super().__init__(data)


class KeyValueWidget(Widget):
    _key = "keyValue"

    def __init__(
        self,
        content: str,
        top: str | None = None,
        bottom: str | None = None,
        break_lines: bool = True,
        on_click: OnClickWidget | None = None,
        icon: str | None = None,
        button: ButtonWidget | None = None,
    ):
        data = dict(
            content=content,
            contentMultiline="true" if break_lines else "false",
        )
        if top:
            data["topLabel"] = top
        if bottom:
            data["bottomLabel"] = bottom
        if on_click:
            data["onClick"] = on_click
        if icon:
            data["icon"] = icon
        if button:
            data["button"] = button

        super().__init__(data)


class TextWidget(Widget):
    _key = "textParagraph"

    def __init__(self, text: str):
        super().__init__(text=text)


class ImageWidget(Widget):
    _key = "image"

    def __init__(self, image_url: str, on_click: OnClickWidget | None = None):
        data = dict(imageUrl=image_url)
        if on_click:
            data.update(on_click.as_data())
        super().__init__(data)


@dataclass
class Section:
    header: str | None = None
    widgets: list[Widget] = field(default_factory=list)

    def add_header(self, text: str):
        self.header = text

    def add_text(
        self,
        content: str,
        title: str = "",
        footer: str = "",
        click_url: str | None = None,
        icon: str | None = None,
        button: str | None = None,
    ):
        widget = KeyValueWidget(
            top=title,
            content=content,
            break_lines=True,
            bottom=footer,
            on_click=OnClickWidget(url=click_url) if click_url else None,
            icon=icon,
            button=button,
        )
        self.widgets.append(widget)

    def add_paragraph(self, text: str):
        self.widgets.append(TextWidget(text=text))

    def add_button(
        self,
        url,
        text: str | None = None,
        image_url: str | None = None,
        icon: str | None = None,
        append: bool = True,
    ):
        button = ButtonWidget(url=url, text=text, image_url=image_url, icon=icon)
        if append and self.widgets and "buttons" in self.widgets[-1]:
            self.widgets[-1]["buttons"].append(button)
        else:
            self.widgets.append(ButtonGroupWidget(buttons=[button]))

    def add_image(self, image_url: str, click_url: str | None = None):
        widget = ImageWidget(
            image_url=image_url,
            on_click=OnClickWidget(url=click_url) if click_url else None,
        )
        self.widgets.append(widget)

    def as_data(self):
        data = {
            "widgets": [widget.as_data() for widget in self.widgets],
        }
        if self.header:
            data["header"] = self.header
        return data

    def __bool__(self):
        return self.header is not None or bool(self.widgets)


@dataclass
class Card:
    header: Widget | None = None
    sections: list[Section] = field(default_factory=list)

    def add_header(self, title: str, subtitle: str = "", image_url: str | None = None, style: str = "IMAGE"):
        self.header = Widget(
            title=title,
            subtitle=subtitle,
            imageUrl=image_url,
            imageStyle=style,
        )

    def add_section(self, section: Section):
        if bool(section):
            self.sections.append(section)

    def as_data(self) -> dict:
        data = {"sections": [section.as_data() for section in self.sections]}
        if self.header:
            data["header"] = self.header.as_data()
        return data


class ChatsHook:
    def __init__(self, hook_url: str):
        self.hook_url = hook_url

    def _post(self, body: dict, thread_key: str | None = None) -> dict:
        url = self.hook_url
        if thread_key:
            url = f"{url}&threadKey={thread_key}"

        response = requests.post(
            url=url,
            headers={"Content-Type": "application/json; charset=UTF-8"},
            data=json.dumps(body),
            timeout=15,
        )
        response.raise_for_status()
        return response.json()

    def send_text(self, text: str, thread_key: str | None = None) -> dict:
        body = {"text": text}
        return self._post(body=body, thread_key=thread_key)

    def send_card(self, card: Card, additional_text: str | None = None, thread_key: str | None = None) -> dict:
        body = {
            "cards": [card.as_data()],
        }

        if additional_text:
            body["text"] = additional_text

        return self._post(body=body, thread_key=thread_key)


class ChatsBot(DiscoveryMixin, GoogleCloudPilotAPI):
    _scopes = ["https://www.googleapis.com/auth/chat.bot"]

    def __init__(self, **kwargs):
        super().__init__(
            serviceName="chat",
            version="v1",
            cache_discovery=False,
            **kwargs,
        )

    def _room_path(self, room_id: str) -> str:
        prefix = "spaces/"
        room_path = f"{prefix}{room_id}" if not room_id.startswith(prefix) else room_id
        return room_path

    def _member_path(self, room_id: str, member_id: str) -> str:
        room_path = self._room_path(room_id=room_id)

        prefix = "members/"
        member_path = f"{prefix}{member_id}" if not member_id.startswith(prefix) else member_id

        return f"{room_path}/{member_path}"

    def get_room(self, room_id: str) -> ResourceType:
        return self._execute(
            method=self.client.spaces().get,
            name=self._room_path(room_id=room_id),
        )

    def get_rooms(self) -> Generator[ResourceType, None, None]:
        yield from self._paginate(
            method=self.client.spaces().list,
            result_key="spaces",
        )

    def get_member(self, room_id: str, member_id: str) -> ResourceType:
        name = self._member_path(room_id=room_id, member_id=member_id)
        return self._execute(
            method=self.client.spaces().members().get,
            name=name,
        )

    def get_members(self, room_id: str) -> Generator[ResourceType, None, None]:
        yield from self._paginate(
            method=self.client.spaces().members().list,
            result_key="memberships",
            params={"parent": self._room_path(room_id=room_id)},
        )

    def send_text(self, room_id: str, text: str) -> ResourceType:
        body = {"text": text}

        return self._execute(
            method=self.client.spaces().messages().create,
            parent=self._room_path(room_id=room_id),
            body=body,
        )

    def send_card(self, room_id: str, card: Card, additional_text: str | None = None) -> ResourceType:
        body = {
            "cards": [card.as_data()],
        }

        if additional_text:
            body["text"] = additional_text

        return self._execute(
            method=self.client.spaces().messages().create,
            parent=self._room_path(room_id=room_id),
            body=body,
        )


__all__ = (
    "Text",
    "Section",
    "Card",
    "ChatsBot",
    "ChatsHook",
)



# Reference <https://cloud.google.com/identity-platform/docs/apis>
import json
import sys
from collections.abc import Iterator
from dataclasses import dataclass
from datetime import UTC, datetime
from enum import Enum
from urllib.parse import parse_qs, urlparse

from google.auth.transport import requests
from google.oauth2 import id_token
from tenacity import Retrying, stop_after_attempt, wait_fixed

from gcp_pilot import exceptions
from gcp_pilot.base import DiscoveryMixin, GoogleCloudPilotAPI, ResourceType


class OOBCodeType(Enum):
    RESET = "PASSWORD_RESET"
    VERIFY = "VERIFY_EMAIL"
    SIGNIN = "EMAIL_SIGNIN"


class FirebaseProviderType(Enum):
    GOOGLE = "google"
    FACEBOOK = "facebook"
    GITHUB = "github"
    MICROSOFT = "microsoft"
    APPLE = "apple"
    PASSWORD = "password"  # TODO: check if this is correct
    PASSWORDLESS = "emailLink"


def parse_timestamp(timestamp: str | int | float) -> datetime | None:
    if not timestamp:
        return None
    if len(str(timestamp)) > 10:
        timestamp = float(timestamp) / 1000
    return datetime.fromtimestamp(float(timestamp), tz=UTC)


@dataclass
class JWTInfo:
    aud: str
    exp: datetime
    iat: datetime
    iss: str
    sub: str

    @property
    def is_expired(self) -> bool:
        return datetime.now(tz=UTC).timestamp() >= self.exp.timestamp()


@dataclass
class User:
    id: str
    email: str
    verified: bool
    disabled: bool
    created_at: datetime | None
    name: str | None = None
    photo_url: bool | None = None
    last_login_at: datetime | None = None
    password_changed_at: datetime | None = None
    extra_attributes: dict[str, str] | None = None
    tenant_id: str | None | None = None

    @classmethod
    def create(cls, data: ResourceType) -> "User":
        return cls(
            id=data["localId"],
            email=data["email"],
            verified=data.get("emailVerified"),
            disabled=data.get("disabled"),
            photo_url=data.get("photoUrl"),
            tenant_id=data.get("tenant_id"),
            created_at=parse_timestamp(timestamp=data["createdAt"]),
            last_login_at=parse_timestamp(timestamp=data.get("lastLoginAt")),
            password_changed_at=parse_timestamp(timestamp=data.get("passwordUpdatedAt")),
            extra_attributes=json.loads(data.get("customAttributes", "{}")),
        )


@dataclass
class FirebaseOAuth:
    id_token: str
    access_token: str
    refresh_token: str | None = None
    token_secret: str | None = None


@dataclass
class FirebaseAuthToken:
    jwt_token: str
    request: requests.Request | None = None
    validate_expiration: bool = True

    def __post_init__(self):
        verification_kwargs = dict(
            id_token=self.jwt_token,
            request=self.request or requests.Request(),
        )
        if not self.validate_expiration:
            verification_kwargs["clock_skew_in_seconds"] = sys.maxsize
        self._data = id_token.verify_firebase_token(**verification_kwargs)

    @property
    def provider_id(self) -> str:
        return self._data["sign_in_method"]

    @property
    def tenant_id(self) -> str | None:
        return self._data.get("tenant_id", None)

    @property
    def oauth(self) -> FirebaseOAuth | None:
        if "oauth_id_token" not in self._data:
            return None

        return FirebaseOAuth(
            id_token=self._data["oauth_id_token"],
            access_token=self._data["oauth_access_token"],
            refresh_token=self._data.get("oauth_refresh_token", None),
            token_secret=self._data.get("oauth_token_secret", None),
        )

    @property
    def event_type(self) -> str:
        return self._data["event_type"]

    @property
    def ip_address(self) -> str:
        return self._data["ip_address"]

    @property
    def user_agent(self) -> str:
        return self._data["user_agent"]

    @property
    def expiration_date(self) -> datetime:
        epoch = self._data["exp"]
        return parse_timestamp(epoch)

    @property
    def user(self) -> User:
        user_data = self._data["user_record"]
        return User(
            id=user_data["uid"],
            email=user_data.get("email"),
            name=user_data.get("display_name"),
            photo_url=user_data.get("photo_url"),
            verified=user_data.get("email_verified"),
            disabled=user_data.get("disabled"),
            tenant_id=user_data.get("tenant_id"),
            created_at=parse_timestamp(user_data["metadata"]["creation_time"]),
            last_login_at=parse_timestamp(user_data["metadata"]["last_sign_in_time"]),
        )

    @property
    def raw_user(self) -> dict:
        return json.loads(self._data["raw_user_info"])

    @property
    def event_id(self) -> dict:
        return self._data["event_id"]

    @property
    def jwt_info(self) -> JWTInfo:
        return JWTInfo(
            aud=self._data["aud"],
            exp=parse_timestamp(timestamp=self._data["exp"]),
            iat=parse_timestamp(timestamp=self._data["iat"]),
            iss=self._data["iss"],
            sub=self._data["sub"],
        )


class IdentityPlatform(DiscoveryMixin, GoogleCloudPilotAPI):
    _scopes = ["https://www.googleapis.com/auth/identitytoolkit"]

    def __init__(self, tenant_id: str | None = None, **kwargs):
        super().__init__(
            serviceName="identitytoolkit",
            version="v1",
            cache_discovery=False,
            static_discovery=False,
            **kwargs,
        )
        self.tenant_id = tenant_id

    def find(
        self,
        email: str | None = None,
        phone_number: str | None = None,
        tenant_id: str | None = None,
        project_id: str | None = None,
    ) -> User:
        try:
            return next(self.lookup(email=email, phone_number=phone_number, tenant_id=tenant_id, project_id=project_id))
        except StopIteration as exc:
            raise exceptions.NotFound() from exc

    def lookup(
        self,
        email: str | None = None,
        phone_number: str | None = None,
        tenant_id: str | None = None,
        project_id: str | None = None,
    ) -> Iterator[User]:
        if not email and not phone_number:
            raise exceptions.ValidationError("Either `email` or `phone_number` must be provided")

        data = {
            "target_project_id": project_id or self.project_id,
            "tenantId": tenant_id or self.tenant_id,
        }
        if email:
            data["email"] = email
        if phone_number:
            if not phone_number.startswith("+"):
                phone_number = f"+{phone_number}"
            data["phone_number"] = phone_number

        response = self._execute(method=self.client.accounts().lookup, body=data)
        for item in response.get("users", []):
            yield User.create(data=item)

    def sign_in_with_password(self, email: str, password: str, tenant_id: str | None = None):
        data = {
            "email": email,
            "password": password,
            "tenantId": tenant_id or self.tenant_id,
        }
        response = self._execute(method=self.client.accounts().signInWithPassword, body=data)
        return response

    def sign_in_with_phone_number(self, phone_number: str, code: str, tenant_id: str | None = None):
        data = {
            "phone_number": phone_number,
            "code": code,
            "tenantId": tenant_id or self.tenant_id,
        }
        response = self._execute(method=self.client.accounts().signInWithPhoneNumber, body=data)
        return response

    def sign_in_with_custom_token(self, token: str, tenant_id: str | None = None):
        data = {
            "token": token,
            "returnSecureToken": True,
            "tenantId": tenant_id or self.tenant_id,
        }
        response = self._execute(method=self.client.accounts().signInWithCustomToken, body=data)
        return response

    def sign_in_with_email_link(self, email: str, code: str, tenant_id: str | None = None):
        data = {
            "email": email,
            "oobCode": code,
            "tenantId": tenant_id or self.tenant_id,
        }
        response = self._execute(method=self.client.accounts().signInWithEmailLink, body=data)
        return response

    def list_users(self, tenant_id: str = None, project_id: str | None = None):
        params = {
            "targetProjectId": project_id or self.project_id,
            "tenantId": tenant_id or self.tenant_id,
        }
        method = self.client.projects().accounts().batchGet
        for item in self._paginate(method=method, result_key="users", pagination_key="nextPageToken", params=params):
            yield User.create(data=item)

    def generate_email_code(
        self,
        type: OOBCodeType,
        email: str,
        ip_address: str | None = None,
        project_id: str | None = None,
        send_email: bool = False,
        redirect_url: str | None = None,
        tenant_id: str | None = None,
    ) -> ResourceType:
        data = {
            "requestType": type.value,
            "email": email,
            "user_ip": ip_address,
            "continue_url": redirect_url,
            "target_project_id": project_id or self.project_id,
            "returnOobLink": not send_email,
            "tenantId": tenant_id or self.tenant_id,
        }
        response = self._execute(method=self.client.accounts().sendOobCode, body=data)

        if send_email:
            return {}

        url = response["oobLink"]
        query = parse_qs(urlparse(url).query)
        return {"url": url, "code": query["oobCode"][0]}

    def reset_password(
        self,
        email: str,
        new_password: str,
        old_password: str = None,
        oob_code: str = None,
        tenant_id: str = None,
    ):
        data = {
            "newPassword": new_password,
            "email": email,
            "tenantId": tenant_id or self.tenant_id,
        }

        if oob_code:
            data["oobCode"] = oob_code
        elif old_password:
            data["oldPassword"] = old_password
        else:
            raise exceptions.ValidationError("Either `old_password` or `oob_code` must be provided")

        response = self._execute(method=self.client.accounts().resetPassword, body=data)
        return response

    def delete_user(self, user_id: str, project_id: str | None = None, tenant_id: str | None = None):
        data = {
            "localId": user_id,
            "targetProjectId": project_id or self.project_id,
            "tenantId": tenant_id or self.tenant_id,
        }
        response = self._execute(method=self.client.accounts().delete, body=data)
        return response

    def disable_user(self, user_id: str, project_id: str | None = None, tenant_id: str | None = None):
        data = {
            "localId": user_id,
            "disableUser": True,
            "targetProjectId": project_id or self.project_id,
            "tenantId": tenant_id or self.tenant_id,
        }
        response = self._execute(method=self.client.accounts().update, body=data)
        return response

    def enable_user(self, user_id: str, project_id: str | None = None, tenant_id: str = None):
        data = {
            "localId": user_id,
            "disableUser": False,
            "targetProjectId": project_id or self.project_id,
            "tenantId": tenant_id or self.tenant_id,
        }
        response = self._execute(method=self.client.accounts().update, body=data)
        return response

    def sign_up(
        self,
        email: str,
        password: str,
        phone_number: str | None = None,
        name: str | None = None,
        photo_url: str | None = None,
        user_id: str | None = None,
        project_id: str | None = None,
        tenant_id: str | None = None,
    ):
        if phone_number and not phone_number.startswith("+"):
            phone_number = f"+{phone_number}"

        data = {
            "email": email,
            "password": password,
            "displayName": name,
            "photoUrl": photo_url,
            "disabled": False,
            "localId": user_id,
            "phoneNumber": phone_number,
            "targetProjectId": project_id or self.project_id,
            "tenantId": tenant_id or self.tenant_id,
        }
        self._execute(method=self.client.accounts().signUp, body=data)

        for attempt in Retrying(stop=stop_after_attempt(3), wait=wait_fixed(1)):
            with attempt:
                return self.find(email=email, project_id=project_id, tenant_id=tenant_id or self.tenant_id)

    def update(
        self,
        user_id: str,
        email: str | None = None,
        password: str | None = None,
        phone_number: str | None = None,
        name: str | None = None,
        photo_url: str | None = None,
        project_id: str | None = None,
        attributes: dict[str, str] | None = None,
        tenant_id: str | None = None,
        enabled: bool | None = None,
    ):
        data = {
            "localId": user_id,
            "targetProjectId": project_id or self.project_id,
            "tenantId": tenant_id or self.tenant_id,
        }
        if enabled:
            data["disableUser"] = not enabled
        if email:
            data["email"] = email
        if password:
            data["password"] = password
        if phone_number:
            if not phone_number.startswith("+"):
                phone_number = f"+{phone_number}"
            data["phoneNumber"] = phone_number
        if name:
            data["displayName"] = name
        if photo_url:
            data["photoUrl"] = photo_url
        if attributes:
            data["customAttributes"] = json.dumps(attributes)
        response = self._execute(method=self.client.accounts().update, body=data)
        return response


class IdentityPlatformAdmin(DiscoveryMixin, GoogleCloudPilotAPI):
    _scopes = ["https://www.googleapis.com/auth/identitytoolkit"]

    def __init__(self, tenant_id: str | None = None, **kwargs):
        super().__init__(
            serviceName="identitytoolkit",
            version="v2",
            cache_discovery=False,
            static_discovery=False,
            **kwargs,
        )
        self.tenant_id = tenant_id

    def _config_path(self, project_id: str | None = None) -> str:
        project_path = self._project_path(project_id=project_id)
        return f"{project_path}/config"

    def get_config(self, project_id: str | None = None) -> dict:
        config_path = self._config_path(project_id=project_id)
        response = self._execute(
            method=self.client.projects().getConfig,
            name=config_path,
        )
        return response

    def add_authorized_domains(self, domains: list[str], project_id: str | None = None) -> dict:
        config = self.get_config(project_id=project_id)
        existing_domains = set(config.get("authorizedDomains", []))
        all_domains = existing_domains.union(set(domains))
        if all_domains != existing_domains:  # only if there's new domains
            return self.set_authorized_domains(domains=list(all_domains), project_id=project_id)
        return config

    def remove_authorized_domains(self, domains: list[str], project_id: str | None = None) -> dict:
        config = self.get_config(project_id=project_id)
        existing_domains = set(config.get("authorizedDomains", []))
        all_domains = existing_domains - set(domains)
        if all_domains != existing_domains:  # only if there's domains to remove
            return self.set_authorized_domains(domains=list(all_domains), project_id=project_id)
        return config

    def set_authorized_domains(self, domains: list[str], project_id: str | None = None) -> dict:
        config_path = self._config_path(project_id=project_id)

        response = self._execute(
            method=self.client.projects().updateConfig,
            name=config_path,
            updateMask="authorizedDomains",
            body={"authorizedDomains": domains},
        )
        return response



from collections.abc import Generator

from gcp_pilot.base import DiscoveryMixin, GoogleCloudPilotAPI, ResourceType


class Datastream(DiscoveryMixin, GoogleCloudPilotAPI):
    def __init__(self, **kwargs):
        super().__init__(
            serviceName="datastream",
            version="v1",
            cache_discovery=False,
            **kwargs,
        )

    def _stream_path(self, stream_name: str, location: str | None = None, project_id: str | None = None) -> str:
        location_path = self._location_path(location=location, project_id=project_id)
        return f"{location_path}/streams/{stream_name}"

    def _object_path(
        self,
        object_id: str,
        stream_name: str,
        location: str | None = None,
        project_id: str | None = None,
    ) -> str:
        stream_path = self._stream_path(stream_name=stream_name, location=location, project_id=project_id)
        return f"{stream_path}/objects/{object_id}"

    def get_streams(
        self,
        location: str | None = None,
        project_id: str | None = None,
    ) -> Generator[ResourceType, None, None]:
        location_path = self._location_path(location=location, project_id=project_id)
        params = {"parent": location_path}
        yield from self._paginate(
            method=self.client.projects().locations().streams().list,
            result_key="streams",
            params=params,
        )

    def get_stream(self, stream_name: str, location: str | None = None, project_id: str | None = None):
        name = self._stream_path(stream_name=stream_name, location=location, project_id=project_id)
        return self._execute(
            method=self.client.projects().locations().streams().get,
            name=name,
        )

    def delete_stream(self, stream_name: str, location: str | None = None, project_id: str | None = None):
        name = self._stream_path(stream_name=stream_name, location=location, project_id=project_id)
        return self._execute(
            method=self.client.projects().locations().streams().delete,
            name=name,
        )

    def get_objects(
        self,
        stream_name: str,
        location: str | None = None,
        project_id: str | None = None,
    ) -> Generator[ResourceType, None, None]:
        stream_path = self._stream_path(stream_name=stream_name, location=location, project_id=project_id)
        params = {"parent": stream_path}
        yield from self._paginate(
            method=self.client.projects().locations().streams().objects().list,
            result_key="streamObjects",
            params=params,
        )

    def get_object(
        self,
        object_id: str,
        stream_name: str,
        location: str | None = None,
        project_id: str | None = None,
    ):
        name = self._object_path(
            object_id=object_id,
            stream_name=stream_name,
            location=location,
            project_id=project_id,
        )
        return self._execute(
            method=self.client.projects().locations().streams().objects().get,
            name=name,
        )

    def find_object(
        self,
        schema: str,
        table: str,
        stream_name: str,
        location: str | None = None,
        project_id: str | None = None,
    ):
        stream_path = self._stream_path(stream_name=stream_name, location=location, project_id=project_id)
        body = {
            "sourceObjectIdentifier": {
                "postgresqlIdentifier": {"schema": schema, "table": table},
            },
        }
        return self._execute(
            method=self.client.projects().locations().streams().objects().lookup,
            parent=stream_path,
            body=body,
        )

    def start_backfill(
        self,
        schema: str,
        table: str,
        stream_name: str,
        location: str | None = None,
        project_id: str | None = None,
    ):
        object_info = self.find_object(
            schema=schema,
            table=table,
            stream_name=stream_name,
            location=location,
            project_id=project_id,
        )
        return self._execute(
            method=self.client.projects().locations().streams().objects().startBackfillJob,
            object=object_info["name"],
        )

    def stop_backfill(
        self,
        schema: str,
        table: str,
        stream_name: str,
        location: str | None = None,
        project_id: str | None = None,
    ):
        object_info = self.find_object(
            schema=schema,
            table=table,
            stream_name=stream_name,
            location=location,
            project_id=project_id,
        )

        return self._execute(
            method=self.client.projects().locations().streams().objects().stopBackfillJob,
            object=object_info["name"],
        )
